/*
 * "Require NetID Authentication" module that allows a server to act as a
 * client of the Central Authentication Service.
 *
 * Original author: Shawn Bayern <shawn.bayern@yale.edu>
 * Original version: February, 2001
 *
 * Current maintainer: Drew Mazurek <drew.mazurek@yale.edu>
 *
 * Note: the "LOG" messages are entirely for debugging and should not be
 * enabled for any production deployment.  They're somewhat arbitrary;
 * we've found them useful at Yale for debugging new features, so I've
 * left them in in case they're useful.
 */

/* $Id: mod_cas2.c,v 1.10 2006-04-08 16:02:58+09 naito Exp $ */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"
#ifdef APACHE2
# include "apr_strings.h"
# include "http_connection.h"
# include "ap_config.h"
#endif /* APACHE2 */

#include <sys/file.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <grp.h>
#include <pwd.h>

#include "cas.h"
#include "ticketcache.h"

#define URLBUFSIZE 4096
#define USERBUFSIZE 20
#define TICKETBUFSIZE 60

// support two different cookie names to allow secure and insecure cookies
// to exist side-by-side
#define COOKIENAME "MODCASID"
#define COOKIENAME_SECURE "MODCASIDS"
#define COOKIEPATH "/"

///////
#ifdef APACHE2
# define RUSER	r->user
#else /* APACHE1 */
# define RUSER	r->connection->user
#endif
///////
#ifdef APACHE2
# define LOCAL_PORT	r->connection->local_addr->port
#else /* APACHE1 */
# define LOCAL_PORT	r->server->port
#endif
///////
#ifdef APACHE2
# define AP_LOG_ERROR(A,B,C)	ap_log_error(A, B, 0, r->server, C)
#else /* APACHE1 */
# define AP_LOG_ERROR(A,B,C)	ap_log_rerror(A, B, r, C)
#endif

/* Utility function declarations */
static char *get_service(request_rec *r, char *buf, int buflen);
static char *get_ticket(request_rec *r);
static char *get_ticket_from_cookies(request_rec *r);
static void log(const char *msg) ;
static void logout(request_rec *r);
static void cache_invalidate(char *ticket);
#ifdef APACHE2
static int is_user_in_group(const char *user, const char *group, apr_pool_t *p);
#else /* APACHE1 */
static int is_user_in_group(const char *user, const char *group, pool *p);
#endif 
static int is_parameter_true(request_rec *r, char *param);
static void write_lock(int fd);
static void read_lock(int fd);
static void un_lock(int fd);
static void make_redirect_login(char *temp, char *code, mod_cas2_conf *c) ;
static void make_redirect_goback(char *temp, char *code, mod_cas2_conf *c) ;

/* Our exported link to Apache. */
/* Our actual module structure */
#ifdef APACHE2
module AP_MODULE_DECLARE_DATA cas2_module; 
#else /* APACHE1 */
module MODULE_VAR_EXPORT cas2_module;
#endif 

/* support functions that we want to prototype here */
static int check_individual_cookie(void *netid, char *key, char *value);
static int check_individual_cookie_for_ticket(void *netid, char *key, char *value);
static char *get_netid_from_cookies(request_rec *r);
static void create_and_send_new_ticket(request_rec *r);

/* effectively per-process local pointer to our mmap()'d ticket cache */
static char *ticket_cache = NULL;
static size_t ticket_cache_size = 0;
static int ticket_cache_fd;

#include "cas_init.c"

/*
 * Called upon the exit of a child process.  mod_cas currently takes this
 * opportunity to munmap() our cache database if one was previously
 * mmap()'d.
 */
#ifdef APACHE2
static apr_status_t cleanup_child(void *server_rec) {
#else /* APACHE1 */
static void cleanup_child(server_rec *s, pool *p) {
#endif 
    if (ticket_cache)
	munmap(ticket_cache, ticket_cache_size);
#ifdef APACHE2    
    return APR_SUCCESS;
#endif
}


/*
 * Called upon the initializaion of a child process.  mod_cas currently
 * takes this opportunity to mmap() our cache database if one has been
 * requested.
 */
#ifdef APACHE2
static void init_child(apr_pool_t *p, server_rec *s) {
#else /* APACHE1 */
static void init_child(server_rec *s, pool *p) {
#endif
    int fd;
    mod_cas2_conf *c;
#ifdef APACHE2
    apr_pool_cleanup_register(p, s, cleanup_child, cleanup_child);
#endif 

    LOG("in init_child()\n");
    // if ticket_cache is already initialized, don't bother re-mapping it
    if (ticket_cache)
	return;

    // retrieve our configuration
    c = (mod_cas2_conf *) ap_get_module_config(s->module_config, &cas2_module);

    if (c != NULL) 
        LOG("-> c is not null\n");
    else
        LOG("-> c is null\n");

    // don't do anything if the server admin doesn't want a cache
    if (!(c->CASLocalCacheFile))
	return;

    LOG("-> dereferenced c\n");

    // if a cache file is requested, mmap() it and store its in-memory
    // location in our static, per-process cell
    fd = open(c->CASLocalCacheFile, O_RDWR | O_CREAT, 0600);
    if (fd < 0) {
      LOG("-> CAS cache file"); LOG(c->CASLocalCacheFile) ; LOG("\n") ;
	LOG("-> couldn't open CAS cache file\n");
	return;					// open() failure
    }

    // make sure the file is as large as we need it to be
    ticket_cache_size =
	sizeof(int) + (c->CASLocalCacheSize * sizeof(TicketEntry));
    lseek(fd, ticket_cache_size, SEEK_SET);
    write(fd, "", 1);

    ticket_cache =
	mmap(0, ticket_cache_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);

    /* 
     * If mmap() failed, there's really not much we can do other than
     * abandon the idea of a ticket cache, at least for us.  But at least
     * we can set ticket_cache back to NULL as a marker.
     */
    if (ticket_cache == MAP_FAILED) {
	ticket_cache = NULL;
	LOG("-> mmap() failed\n");
	return;
    }

    ticket_cache_fd = fd;

    LOG("-> init_child() done\n");
}


/*
 * Handles CAS authentication.  Redirects to CAS if we can't authenticate
 * the user, and expects a ticket back.  Sets r->user with
 * a validated user.  Operates only when the AuthType is set to "CAS."
 *
 * Once past this point, we never directly return DECLINED; we return
 * success with OK and, if we can't authenticate the user, merely redirect
 * to the CAS with HTTP_MOVED_TEMPORARILY.  On error, we return
 * HTTP_INTERNAL_SERVER_ERROR.  If we return OK, no other modules are
 * supposed to handle this request during the "check user" phase.
 */
static int do_cas(request_rec *r)
{
    // it's okay to use these buffers since we're single-threaded
    char *urlbuf = (char *) apr_pcalloc(r->pool, URLBUFSIZE);
    char *userbuf = (char *) apr_pcalloc(r->connection->pool, USERBUFSIZE);
    char *service, *ticket;
    const char *auth_type = ap_auth_type(r);
    char *p;
    int flag=0 ;
    int validate_status = 0;
    int (*CAS_validate)(char *, char *, char *, int, mod_cas2_conf *);
    int CAS_validate_SSL(char *, char *, char *, int, mod_cas2_conf *);
    int CAS_validate_non_SSL(char *, char *, char *, int, mod_cas2_conf *);
    char *cas_fail_code[] = {"", "", "INVALID_SERVICE", "INVALID_TICKET", "ACCESS_DENIED"} ;
    
    mod_cas2_conf *c = (mod_cas2_conf *) ap_get_module_config(
	r->server->module_config, &cas2_module);

    LOG("in do_cas()\n");

    /* We only have something to do if "CAS" is the AuthType. */
    if (auth_type && strcmp(auth_type, "CAS")) {
      LOG("out do_cas() by DECLINED") ;
      return DECLINED;
    }


    /*
     * Okay, we're relevant.  Now, if we've been handed a cookie that
     * identifies the user (with the help of the ticket cache), we can
     * simply honor the user's preauthenticated status.  (This call
     * uses the ticket cache behind the scenes.  Separately, as a nice
     * side effect, this sets r->user to NULL if we don't
     * find a cookie; that way, if we proceed, we know that r->connnection->
     * user doesn't have some bogus value in in.
     */

    if (RUSER = get_netid_from_cookies(r)) {
        mod_cas2_conf *c;

        // retrieve our configuration
        c = (mod_cas2_conf *) ap_get_module_config(
	    r->server->module_config, &cas2_module);

	if (c->CASLogoutParameter && c->CASLogoutLocalURL 
		&& is_parameter_true(r,c->CASLogoutParameter)) {
	    LOG("-> logout parameter set... logging user out\n");
	    logout(r);
	    LOG("-> redirecting to logout page\n");
	    apr_table_set(r->headers_out, "Location", c->CASLogoutLocalURL);
	    return HTTP_MOVED_TEMPORARILY;
	} else {
	    LOG("-> authenticated from cookie\n");
	    return OK;
	}
    }

    /*
     * "CAS" is the AuthType, so we consider the request in terms of
     * a service and a ticket.
     */
    service = get_service(r, urlbuf, URLBUFSIZE);
    ticket = get_ticket(r);

    LOG("  service = '"); LOG(service);
    LOG("'; ticket = '");
    if (ticket)
        LOG(ticket);
    else
        LOG("(null)");
    LOG("'\n");

    if (service == NULL) {
      AP_LOG_ERROR(APLOG_MARK, APLOG_ERR, "mod_cas: null " "service");
      return HTTP_INTERNAL_SERVER_ERROR;	/* shouldn't happen */
    }

    if (c->CASValidateSSL) {
      LOG("CAS_validate is SSL\n") ;
      CAS_validate = CAS_validate_SSL ;
    }
    else {
      LOG("CAS_validate is NON SSL\n") ;
      CAS_validate = CAS_validate_non_SSL ;
    }

    LOG("about to call CAS_validate()\n");

    if (ticket == NULL) LOG("TICKET is NULL\n") ;
    else LOG("TICKET is NON NULL\n") ;

    // INVALID_TICKET λˤ login  redirect
    // ¾λˤ goBack 
    // http://CAS/cas/goBack?edu.yale.its.tp.cas.status=$CODE  redirect 
    
    if (ticket != NULL
        && (validate_status = CAS_validate(ticket, service, userbuf, USERBUFSIZE, c)) == CAS_SUCCESS) {
/*        log("-> successful primary authentication for '"); log(userbuf); log("'\n");  */
      // Successful authentication

    {
      require_line *requirements;
#ifdef APACHE2
      const apr_array_header_t *requirements_array;
#else /* APACHE1 */
      const array_header *requirements_array;
#endif
      const char *req_line, *req_word;
      int i, method; 

      requirements_array = ap_requires(r);
      requirements = (require_line *) requirements_array->elts;
      method = r->method_number;
      for (i = 0; i < requirements_array->nelts; i++) {
	if (!(requirements[i].method_mask & (1 << method))) continue ; 
	req_line = requirements[i].requirement;		// full line
	req_word = ap_getword(r->pool, &req_line, ' ');	// individual word
	/* valid-user Τߤʤ OK ֤ (flag = 0 Ǹ夫 OK Ȥ)
	 * ʳ word äƤ, user  check
	 */
	if (strcmp(req_word, "valid-user")==0) {
/*  	  log("-> req_word ") ; log(req_word); log("\n");  */
	  while (*req_line) {
	    flag = 1 ;
	    req_word = ap_getword_conf(r->pool, &req_line);
/*  	    log("-> req_word ") ; log(req_word); log("\n");  */
	    if (strcmp(req_word, userbuf) == 0) return OK ;
	  }
	}
      }
    }
    /*
    {
      char t[100] ;
      sprintf(t, "flag = %d\n", flag) ;
      log(t) ;
    }
    */
    if (flag == 0) return OK ;
    {
      char *temp = alloca(strlen("https://") + strlen(c->CASHost) + strlen("/") + 50) ;
      make_redirect_goback(temp, cas_fail_code[validate_status], c) ;
      apr_table_set(r->headers_out, "Location", temp); 
      return HTTP_MOVED_TEMPORARILY;
    }
    } else if (validate_status == BAD_CERT) {
      LOG("-> certificate error\n");
      AP_LOG_ERROR(APLOG_MARK, APLOG_ERR, "mod_cas: certificate " "error: bad cert.");
      return HTTP_INTERNAL_SERVER_ERROR;
    } else if ((ticket != NULL)&&(validate_status != CAS_FAIL_INVALID_TICKET)) {
      LOG("-> certificate error:\n") ; 
      char *temp = alloca(strlen("https://") + strlen(c->CASHost) + strlen("/") + 50) ;
      if (!temp) {
	AP_LOG_ERROR(APLOG_MARK, APLOG_ERR, "mod_cas: alloca " "failed in function create_and_send_new_ticket()");
	return HTTP_INTERNAL_SERVER_ERROR;
      }
      make_redirect_goback(temp, cas_fail_code[validate_status], c) ;
      apr_table_set(r->headers_out, "Location", temp);
      return HTTP_MOVED_TEMPORARILY ;
    } else {
      /* Bad or missing ticket, so bounce the user to CAS. */
      /* "10" is for PORT NUMBER */
      char *temp = alloca(strlen("https://") + 
          strlen(c->CASHost) + strlen("/") + strlen(c->CASLogin) + strlen("?service=")
          + strlen(service) + 10);
      if (!temp) {
	AP_LOG_ERROR(APLOG_MARK, APLOG_ERR, "mod_cas: alloca " "failed in function create_and_send_new_ticket()");
	return HTTP_INTERNAL_SERVER_ERROR;
      }
      make_redirect_login(temp, service, c) ;
      apr_table_set(r->headers_out, "Location", temp);
      LOG("SERVICE ") ; LOG(temp) ; LOG("\n") ;
      return HTTP_MOVED_TEMPORARILY;
    }
}

static void make_redirect_login(char *temp, char *service, mod_cas2_conf *c) 
{
      if (c->CASLoginSSL) strcpy(temp, "https://" ) ;
      else strcpy(temp,"http://") ;
      strcat(temp, c->CASHost);
      if (((c->CASLoginSSL)&&(strcmp(c->CASPort, "443") != 0))||((!c->CASLoginSSL)&&(strcmp(c->CASPort, "80") != 0))) 
      {
	strcat(temp,":") ; strcat(temp,c->CASPort) ;
      }
      strcat(temp, c->CASLogin) ;
      strcat(temp, "?service=");
      strcat(temp, service);
      return ;
}

// http://CAS/cas/goBack?edu.yale.its.tp.cas.status=$CODE  redirect 
static void make_redirect_goback(char *temp, char *code, mod_cas2_conf *c) 
{
      if (c->CASGobackSSL) strcpy(temp, "https://" ) ;
      else strcpy(temp,"http://") ;
      strcat(temp, c->CASHost);
      if (((c->CASGobackSSL)&&(strcmp(c->CASPort, "443") != 0))||((!c->CASGobackSSL)&&(strcmp(c->CASPort, "80") != 0))) 
      {
	strcat(temp,":") ; strcat(temp,c->CASPort) ;
      }
      strcat(temp, c->CASGoback) ;
      strcat(temp, "?edu.yale.its.tp.cas.status=");
      strcat(temp, code);
      LOG("REDIRECT: ") ; LOG(temp) ; LOG("\n") ;
      return ;
}

/*
 * Compares the authenticated user (from CAS) with the list of users (and
 * groups) authorized by the "Require" directives.  NOTE: We implement
 * set union here because that's how Apache modules typically work; the word
 * "Require" might incorrectly lead you to assume that we perform set
 * intersection.  For instance, "Require user bayern; Require group foo"
 * means "let anyone if if they're either user bayern or in group foo."  
 * All user authentication proceeds through CAS; group authentication maps
 * a CAS-authenticated username against the machine's local groups.
 * (In the future, this could change to support a more generalized
 * framework for group discovery.)
 *
 * Operates only if the AuthType is CAS, returning DECLINED otherwise on
 * the chance that another module will validate the request if we can't.
 * On success, we return OK, which solves the "authorization" question
 * for the entire request -- on OK, other modules should not be called.
 * If the AuthType is CAS, though, and the user isn't authorized as far
 * as we're concerned, we're going to take the initiative and return
 * HTTP_FORBIDDEN directly (rather than just DECLINED; that is, we're
 * not going to give any other module a chance to authorize anyone if the
 * AuthType is CAS).  We return HTTP_FORBIDDEN rather than
 * HTTP_UNAUTHORIZED because we don't want the browser to think it's
 * just temporarily failed and conduct a Basic Authentication dialog with
 * the user.
 */
static int authorize(request_rec *r)
{
#ifdef APACHE2
    const apr_array_header_t *requirements_array;
#else /* APACHE1 */
    const array_header *requirements_array;
#endif
    const char *auth_type = ap_auth_type(r);
    const char *authenticated_user = RUSER;
    const char *req_line, *req_word;
    require_line *requirements;
    int i, method;
    

    LOG("in authorize()\n");
    return DECLINED;
    LOG("authtype=") ; LOG(auth_type) ; LOG("\n");

    /* Someone else has to pick up the torch if "CAS" isn't our AuthType. */
    if (auth_type && strcmp(auth_type, "CAS"))
      return DECLINED;

    LOG("-> AuthType CAS detected\n");
    if (authenticated_user) {
      LOG("-> authenticated username: "); LOG(authenticated_user); LOG("\n");
    } else {
      LOG("-> ... but no authenticated user!  (Shouldn't happen.)\n");
    }

    /* Some design considerations here were informed by mod_auth_kerb */

    requirements_array = ap_requires(r);
    if (!requirements_array)
	return OK;				// no restriction

    requirements = (require_line *) requirements_array->elts;
    method = r->method_number;
    for (i = 0; i < requirements_array->nelts; i++) {

	// if our method isn't covered by the requirements line, ignore it
	if (!(requirements[i].method_mask & (1 << method))) {
	    continue;
	}

	req_line = requirements[i].requirement;		// full line
	req_word = ap_getword(r->pool, &req_line, ' ');	// individual word
	LOG("->req_word ") ; LOG(req_word) ; LOG("\n") ;

	/*
	 * If the requirements simply call for a "valid user," we can
	 * simply return success, because if we're called, it means that
	 * CAS authentication performed in an earlier callback DIDN'T
	 * redirect back to CAS, so we have a valid user already, either
	 * by virtue of a ticket or of having stored authenticated status
	 * earlier.
	 */
	if (!strcmp(req_word, "valid-user"))
	    return OK;
	else if (!strcmp(req_word, "user")) {
	    // the rest of the words are individual user names, so see
	    // if our authenticated user matches
	    while (*req_line) {
		req_word = ap_getword_conf(r->pool, &req_line);
		if (!strcmp(req_word, authenticated_user))
		    return OK;
	    }
	} else if (!strcmp(req_word, "group")) {
	    // the rest of the words are individual group names
	    while (*req_line) {
		req_word = ap_getword_conf(r->pool, &req_line);
		if (is_user_in_group(authenticated_user, req_word, r->pool))
		    return OK;
	    }
	}
    }

    // For now, no other sorts of requirements are supported
    // For the moment, fail cleanly if we get something like
    // "Require group foo" so that configuration files don't
    // need to be changed; the relevant line can just be ignored,
    // and we failed "closed."
    return HTTP_FORBIDDEN;
}

/* Utility function definitions */

/*
 * Fills buf with the 'service' portion of the URL requested.  This will
 * be the entire URL minus any trailing '&ticket=', if present.
 * The URL is determined here, heuristically, from the request record.
 * If buf isn't large enough to fit the entire URL, NULL is returned.
 */
static char *get_service(request_rec *r, char *buf, int buflen) {
    int l;
    char portstr[6];
    char *path = r->uri, *p;
    mod_cas2_conf *c = (mod_cas2_conf *) ap_get_module_config(
	r->server->module_config, &cas2_module);

    // /* 
    //  * For the server name, use the "Host" header if it exists; otherwise,
    //  * consult the 'server' record.  Note that our 'server' string includes
    //  * the server's name and an optional ':port'.
    //  */
    // const char *server = apr_table_get(r->headers_in, "Host");
    // char *query = r->args;
    // char *scheme = "http://";
    // if (!server)
    //	  server = r->server->server_hostname;
    // if (!query)
    //	  query = "";

    /* Avoid using untrusted "Host" header. */
    const char *server = r->server->server_hostname;
    char *query = r->args;
    char *scheme = "http://";
    if (!query)
      query = "";

    if (!(c->CASLocalCacheInsecure))
	scheme = "https://";

    /* convert the port number to a string in case we need it */
    snprintf(portstr, 6*sizeof(char), "%d", LOCAL_PORT);

    LOG("portstr: ");
    LOG(portstr);
    LOG("\n");

    if ((l = strlen(scheme) + strlen(server) + strlen(":") 
	+ strlen(portstr) + strlen(path) + strlen(query) + 1) > buflen)
	return NULL;

    strcpy(buf, scheme);
    strcat(buf, server);

    /* add the port number if it's non-standard */
    if(c->CASLocalCacheInsecure && LOCAL_PORT != 80
	|| !c->CASLocalCacheInsecure && LOCAL_PORT != 443) {
	strcat(buf, ":");
	strcat(buf, portstr);
    }

    strcat(buf, path);
    if (query && *query != '\0') {
	strcat(buf, "%3F");             /* "?" */
        for (p = query; *p; p++) {
            if (*p == '=')
                strcat(buf, "%3D");
            else if (*p == '?')
                strcat(buf, "%3F");
            else if (*p == '&')
                strcat(buf, "%26");	/* rewrite later to be more complete */
            else {
		char t[2];
                t[0] = *p;
                t[1] = '\0';
		strcat(buf, t);
            }
        }
    }

    /* eliminate the ticket */
    if ((p = strstr(buf, "%3Fticket%3D")))
	*p = '\0';
    if ((p = strstr(buf, "%26ticket%3D")))
	*p = '\0';

    return buf;
}

/*
 * Returns the 'ticket' portion of the URL requested.  This consists
 * of everything following '&ticket=' in the logical full URL.  If no "ticket"
 * parameter was specified, we return NULL.
 */
static char *get_ticket(request_rec *r) {
    char *ticket = NULL;

    if (r->args && *(r->args) != '\0') {
	/* if the query starts with "ticket=", then use the whole query */
	if (r->args == strstr(r->args, "ticket="))
	    ticket = r->args + strlen("ticket=");
	else {
	    /* otherwise, we need an "&" */
	    ticket = strstr(r->args, "&ticket=");
	    if (ticket)
		ticket += strlen("&ticket=");
	}
    }

    if (!ticket || *ticket == '\0')
	return NULL;
    return ticket;
}

static int is_parameter_true(request_rec *r, char *param) {
    char *temp;
    char *p;

    if(r->args && *(r->args) != '\0') {
	temp = alloca(strlen(param) + strlen("=true") + 1);
	strcpy(temp, param);
	strcat(temp, "=true");
	if((p = strstr(r->args,temp)) && (p[strlen(temp)] == '&'
	    || (p[strlen(temp)] == '\0'))) {
	    return 1;
	}
    }

    return 0;
}

/* Functions for handling the ticket cache. */

/*
 * Add an entry to the ticket cache.  Values are copied from the TicketEntry,
 * which need not persist after the call terminates.
 */
static void cache_put(TicketEntry te) {
    int next;
    char *last_slot = ticket_cache + ticket_cache_size - sizeof(TicketEntry);

    if (!ticket_cache)
	return;			// alas, nothing to do

    LOG("in cache_put()\n");
    // get an advisory, exclusive lock on the ticket cache file
    write_lock(ticket_cache_fd);

    LOG("-> got lock\n");
    // the first int of the ticket cache stores the next slot
    memcpy(&next, ticket_cache, sizeof(int));

    LOG("-> first memcpy() done\n");
    // wrap around if necessary
    if (ticket_cache + sizeof(int) + next * sizeof(TicketEntry) > last_slot)
	next = 0;
    memcpy(ticket_cache + sizeof(int) + next * sizeof(TicketEntry),
	&te, sizeof(TicketEntry));
    msync(ticket_cache + sizeof(int) + next * sizeof(TicketEntry),
	sizeof(TicketEntry), MS_SYNC | MS_INVALIDATE);

    LOG("-> second memcpy() done\n");
    // increment next, letting our successor handle wraparound
    next++;
    memcpy(ticket_cache, &next, sizeof(int));
    msync(ticket_cache, sizeof(int), MS_SYNC | MS_INVALIDATE);

    // release our lock
    un_lock(ticket_cache_fd);
    LOG("-> released lock\n");
}

/*
 * Marks the given ticket as invalid, effectively logging the user out
 * of Apache.  Hopefully.
 */
static void cache_invalidate(char *ticket) {
    char *last_slot = ticket_cache + ticket_cache_size - sizeof(TicketEntry);
    char *p;
    TicketEntry t;

    LOG("in cache_invalidate()\n");

    if(!ticket_cache) {
	LOG("-> no ticket cache, returning\n");
	return;
    }

    // get an advisory, shared file lock on the ticket cache file
    read_lock(ticket_cache_fd);

    for (p = ticket_cache + sizeof(int);
	    p <= last_slot;
	    p += sizeof(TicketEntry)) {
	memcpy(&t, p, sizeof(TicketEntry));
	if (!strcmp(t.ticket, ticket)) {
	    LOG("-> found ticket entry, invalidating it\n");
	    t.valid = 0;
	    write_lock(ticket_cache_fd);
	    memcpy(p, &t, sizeof(TicketEntry));
	    msync(p, sizeof(TicketEntry), MS_SYNC | MS_INVALIDATE);
	    un_lock(ticket_cache_fd);
	    return;
	}
    }

    un_lock(ticket_cache_fd);

    LOG("-> never found the ticket... going home broke and hungry\n");
}

/*
 * Retrieves the NetID for a given ticket, storing it in the buffer
 * provided.  Returns 1 if the ticket was found and valid, 0 otherwise.
 * The 'insecure' parameter stores our current CASLocalCacheInsecure
 * status.
 */
static int cache_get(char *ticket, char *buf, int buflen, char insecure) {
    char *p;
    char *last_slot = ticket_cache + ticket_cache_size - sizeof(TicketEntry);
    TicketEntry t;

    LOG("in cache_get()\n");
    // get an advisory, shared file lock on the ticket cache file
    read_lock(ticket_cache_fd);

    // the first few bytes of the ticket array form an integer we don't
    // care about
    for (p = ticket_cache + sizeof(int); 
	    p <= last_slot;
	    p += sizeof(TicketEntry)) {
	memcpy(&t, p, sizeof(TicketEntry));
	if (t.valid 
		&& t.secure == !insecure
		&& time(NULL) < t.expiration
		&& !strcmp(t.ticket, ticket)) {
	    if (strlen(t.netid) + 1 > buflen)
		return 0;
	    strcpy(buf, t.netid);
	    un_lock(ticket_cache_fd);			// release the lock
	    return 1;
	}
    }
    un_lock(ticket_cache_fd);				// release the lock
    return 0;
}

/*
 * Constructs a random ticket value and stores it in the buffer provided.
 * The buffer is terminated with '\0'.  On unexpected failure, return 0;
 * otherwise, return 1.
 */
static int random_ticket(char *buf, int buflen, char *egdfile) {
    int fd;

    char *pool =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz"
        "01234567890";

    LOG("in random_ticket()\n");

    if (egdfile == NULL) {
	LOG("using /dev/urandom\n");
        /* Get our randomness from /dev/urandom. */

        if ((fd = open("/dev/urandom", O_RDONLY)) < 0) {
            LOG("ERROR!  Couldn't open /dev/urandom.  Do not deploy!");
            return 0;
	} else if (read(fd, buf, buflen - 1) < 0) {
            close(fd);
            LOG("ERROR!  Couldn't open /dev/urandom.  Do not deploy!");
            return 0;
        } else {
            int i;
	    LOG("processing randomness...\n");
            close(fd);
            buf[buflen - 1] = '\0';
            /* now, take each byte and convert it to a readable one */
            for (i = 0; i < buflen - 1; i++)
                buf[i] = pool[ buf[i] % strlen(pool) ];
	    LOG("got random ticket: ");
	    LOG(buf);
	    LOG("\n");
            return 1;
        }
    } else {
      int i;
      LOG("using EGD\n");

      /** Draw entropy from the EGD if necessary. */
      if (!RAND_status())
          RAND_egd(egdfile);

      /** Now, use OpenSSL to produce random bytes. */
      RAND_pseudo_bytes(buf, buflen -1);
      buf[buflen - 1] = '\0';

      /* Convert each byte to a readable character. */
      for (i = 0; i < buflen - 1; i++)
          buf[i] = pool[ buf[i] % strlen(pool) ];
      return 1;

    }
}

/*
 * Sets a cookie corresponding to a particular ticket in the given
 * request's response headers.  Uses the given configuration record
 * to determine whether the ticket can be sent back to non-SSL URLs.
 */
static void send_cookie(char *ticket, request_rec *r) {
    mod_cas2_conf *c = (mod_cas2_conf *) ap_get_module_config(
	r->server->module_config, &cas2_module);
    char *cookie = apr_pstrcat(r->pool,
        (c->CASLocalCacheInsecure ? COOKIENAME : COOKIENAME_SECURE),
        "=", ticket,
	(c->CASLocalCacheInsecure ? "" : ";Secure"),
	";Path=", COOKIEPATH, NULL);

    LOG("in send_cookie()\n");
    LOG("-> sending cookie '");
    LOG(cookie);
    LOG("'\n");
#ifdef APACHE2
    apr_table_add(r->headers_out, "Set-Cookie", cookie);
#else 
    ap_table_add(r->headers_out, "Set-Cookie", cookie);
#endif 
}

/*
 * Walks through the given request's headers, reading all cookies that
 * can correspond to tickets and checking them against our ticket cache.
 * If a valid NetID is found, we stop immediately and return it;
 * otherwise, if no valid NetID is ever found, we return NULL.
 */
static char *get_netid_from_cookies(request_rec *r) {
    // msg_buf is used as a two-way message buffer between apr_table_do
    // and the function we have it call here.
    // as input to check_individual_cookie:
    //   -- its first byte is our config's value of CASLocalCacheInsecure
    // as output from check_individual_cookie:
    //   -- it stores a NUL-terminated string representing the found
    //      NetID, or is equivalent to "" if no NetID was found.
    char msg_buf[USERBUFSIZE];
    mod_cas2_conf *c;

    c = (mod_cas2_conf *) ap_get_module_config(
	r->server->module_config, &cas2_module);
    msg_buf[0] = c->CASLocalCacheInsecure;		// initialize input
    msg_buf[1] = '\0';

    // fail trivially if there's no ticket cache
    if (!ticket_cache)
	return NULL;

    // the following call ends up with a valid netid or NULL in 'preauth_netid'
    apr_table_do(
	(int (*)()) check_individual_cookie,
	msg_buf,
	r->headers_in,
	"Cookie",
	NULL);

    if (msg_buf[0] == c->CASLocalCacheInsecure
	    && msg_buf[1] == '\0')
	// the buffer is unchanged, meaning nobody wrote a NetID to it
	return NULL;
    else
#ifdef APACHE2
        return apr_pstrdup(r->pool, msg_buf);
#else
	return ap_pstrdup(r->pool, msg_buf);
#endif
}

/*
 * A support function for get_netid_from_cookies, used by apr_table_do
 * to handle individual cookies.  (There may be multiple "Cookie" headers
 * in the request that we receive, so we can't just use ap_table_get.)
 */
static int check_individual_cookie(void *msg_buf, char *key, char *value) {
    char *ticket, *p;
    char netid_buf[USERBUFSIZE];

    LOG("in check_individual_cookie()\n");
    // we're looking for whatever's between COOKIENAME= and ';'
    // (or the end), ignoring optional quotation marks at the beginning or
    // the end
    ticket = alloca(strlen(value) + 1);
    strcpy(ticket, value);

    // msg_buf[0] stores our "insecure" status
    if (*((char *) msg_buf)) {
      ticket = strstr(ticket, COOKIENAME "=");           // find key
      if (!ticket)
          return 1;
      ticket += strlen(COOKIENAME "=");                  // skip key and '='
    } else {
      ticket = strstr(ticket, COOKIENAME_SECURE "=");    // find key
      if (!ticket)
          return 1;
      ticket += strlen(COOKIENAME_SECURE "=");           // skip key and '='
    }

    if (*ticket == '"')
	ticket++;			// if it starts with '"', skip '"'
    p = strchr(ticket, ';');
    if (p)
	*p = '\0';			// end the string at ';' if it's there
    p = ticket + strlen(ticket);
    if (*p == '"')
	*p = '\0';			// if we ended with '"', delete it

    // now, we have the ticket from the cookie.
    // let's see what it buys us.
    // (msg_buf[0] stores our current CASLocalCacheInsecure status)
    if (cache_get(ticket, netid_buf, sizeof(netid_buf), *((char *) msg_buf))) {
	// copy the netid back out
	// the 'msg_buf' parameter is a char[] of size USERBUFSIZE,
	// so it can just take the NetID we found, byte for byte
	strcpy((char *) msg_buf, netid_buf);

        LOG("-> check_individual_cookie succeeded; found ");
        LOG(netid_buf);
        LOG("\n");

	return 0;			// stop processing in apr_table_do
    }
    return 1;				// continue process in apr_table_do
}

static void logout(request_rec *r) {
    char *ticket;

    LOG("in logout()\n");

    // if there's no ticket cache, there's nothing to do
    if(!ticket_cache) {
	LOG("-> no ticket cache... don't need to logout\n");
	return;
    }

    ticket = get_ticket_from_cookies(r);
    cache_invalidate(ticket);

    return;
}
    
static char *get_ticket_from_cookies(request_rec *r) {
    // msg_buf is used as a two-way message buffer between apr_table_do
    // and the function we have it call here.
    // as input to check_individual_cookie:
    //   -- its first byte is our config's value of CASLocalCacheInsecure
    // as output from check_individual_cookie:
    //   -- it stores a NUL-terminated string representing the found
    //      NetID, or is equivalent to "" if no NetID was found.
    char msg_buf[TICKETBUFSIZE];
    mod_cas2_conf *c;

    c = (mod_cas2_conf *) ap_get_module_config(
        r->server->module_config, &cas2_module);
    msg_buf[0] = c->CASLocalCacheInsecure;              // initialize input
    msg_buf[1] = '\0';

    // fail trivially if there's no ticket cache
    if (!ticket_cache)
        return NULL;

    // the following call ends up with a valid netid or NULL in 'preauth_netid'
    apr_table_do(
        (int (*)()) check_individual_cookie_for_ticket,
        msg_buf,
        r->headers_in,
        "Cookie",
        NULL);

    if (msg_buf[0] == c->CASLocalCacheInsecure
            && msg_buf[1] == '\0')
        // the buffer is unchanged, meaning nobody wrote a ticket to it
        return NULL;
    else
#ifdef APACHE2
        return apr_pstrdup(r->pool, msg_buf);
#else
	return ap_pstrdup(r->pool, msg_buf);
#endif
}

/*
 * A support function for get_netid_from_cookies, used by apr_table_do
 * to handle individual cookies.  (There may be multiple "Cookie" headers
 * in the request that we receive, so we can't just use ap_table_get.)
 */
static int check_individual_cookie_for_ticket(void *msg_buf, char *key, char *value) {
    char *ticket, *p;

    LOG("in check_individual_cookie_for_ticket()\n");
    // we're looking for whatever's between COOKIENAME= and ';'
    // (or the end), ignoring optional quotation marks at the beginning or
    // the end
    ticket = alloca(strlen(value) + 1);
    strcpy(ticket, value);

    // msg_buf[0] stores our "insecure" status
    if (*((char *) msg_buf)) {
      ticket = strstr(ticket, COOKIENAME "=");           // find key
      if (!ticket)
          return 1;
      ticket += strlen(COOKIENAME "=");                  // skip key and '='
    } else {
      ticket = strstr(ticket, COOKIENAME_SECURE "=");    // find key
      if (!ticket)
          return 1;
      ticket += strlen(COOKIENAME_SECURE "=");           // skip key and '='
    }

    if (*ticket == '"')
	ticket++;			// if it starts with '"', skip '"'
    p = strchr(ticket, ';');
    if (p)
	*p = '\0';			// end the string at ';' if it's there
    p = ticket + strlen(ticket);
    if (*p == '"')
	*p = '\0';			// if we ended with '"', delete it

    strcpy((char *) msg_buf, ticket);

    LOG("-> check_individual_cookie_for_ticket succeeded; found ");
    LOG(ticket);
    LOG("\n");

    return 0;			// stop processing in apr_table_do
}

/*
 * Creates and sends a cookie holding a new ticket, which we add to
 * the ticket cache under the appropriate NetID.
 */
static void create_and_send_new_ticket(request_rec *r) {
    mod_cas2_conf *c;
    TicketEntry t;

    LOG("in create_and_send_new_ticket()\n");
    c = (mod_cas2_conf *) ap_get_module_config(
	r->server->module_config, &cas2_module);

    // don't do anything if we're not supposed to have a ticket cache,
    // or if the ticket cache doesn't exist
    if (!c->CASLocalCacheFile || !ticket_cache)
	return;

    // set up the TicketEntry
    t.valid = 1;
    t.secure = !(c->CASLocalCacheInsecure);
    strcpy(t.netid, RUSER);
    if (random_ticket(t.ticket, sizeof(t.ticket), c->CASEGDFile)) {
	LOG("random_ticket successful: ticket = ");
	LOG(t.ticket);
	LOG("\n");
        t.expiration = time(NULL) + c->CASLocalCacheTimeout;
        cache_put(t);
        send_cookie(t.ticket, r);
    }
}

/* Returns 1 if 'user' is in 'group', 0 otherwise. */
static int is_user_in_group(const char *user,
			    const char *group,
#ifdef APACHE2
			    apr_pool_t *p
#else
			    pool *p
#endif
) {
    // TODO: cache here for performance?  (it's cleaner to cache in nscd)
    char *gn;
    char **t;
    struct group *g;
    struct passwd *u;

    // this isn't thread-safe, but the current APR doesn't support generalized
    // access to getgrnam() or getpwnam()

    if (!(g = getgrnam(group)))
        return 0;
    for (t = g->gr_mem; t && *t; t++) {
	LOG("checking "); LOG(user); LOG(" against"); LOG(*t); LOG("\n");
        if (!strcmp(*t, user))
            return 1;
    }
    if (!(u = getpwnam(user)))
        return 0;
#ifdef APACHE2
    if (apr_group_name_get(&gn, u->pw_gid, p) != APR_SUCCESS)
        return 0;
#else
    if (!(g = getgrgid(u->pw_gid)))
        return 0;
#endif 
    if (!strcmp(gn, group))
	return 1;
    else
	return 0;
}

static void write_lock(int fd) {
    /* prepare for writelock on (or unlock of) entire file */
    struct flock lock;
    lock.l_type = F_WRLCK;
    lock.l_whence = 0;
    lock.l_start = 0L;
    lock.l_len = 0L;

    /* block until we can get a shared lock */
    fcntl(fd, F_SETLKW, &lock);
}

static void read_lock(int fd) {
    /* prepare for writelock on (or unlock of) entire file */
    struct flock lock;
    lock.l_type = F_RDLCK;
    lock.l_whence = 0;
    lock.l_start = 0L;
    lock.l_len = 0L;

    /* block until we can get a shared lock */
    fcntl(fd, F_SETLKW, &lock);
}

static void un_lock(int fd) {
    /* prepare for writelock on (or unlock of) entire file */
    struct flock lock;
    lock.l_type = F_UNLCK;
    lock.l_whence = 0;
    lock.l_start = 0L;
    lock.l_len = 0L;

    /* block until we can get a shared lock */
    fcntl(fd, F_SETLKW, &lock);
}

static void log(const char *msg) { 
  FILE *fp ;
  fp = fopen("/tmp/log", "a");
  fputs(msg, fp);
  fflush(fp);
  fclose(fp) ;
}

#include "cas_directive.c"


#ifdef APACHE2
/* Apache 2.0-style hook registration */
static void cas_register_hooks(apr_pool_t *p) {
  ap_hook_check_user_id(do_cas, NULL, NULL, APR_HOOK_MIDDLE);
/*    ap_hook_auth_checker(authorize, NULL, NULL, APR_HOOK_MIDDLE); */
  ap_hook_child_init(init_child, NULL, NULL, APR_HOOK_MIDDLE);
}
#endif

/* Our actual module structure */
#ifdef APACHE2
module AP_MODULE_DECLARE_DATA cas2_module =
{
  STANDARD20_MODULE_STUFF,		/* header */
  NULL,				/* per-directory init */
  NULL,				/* per-directory merge */
  init_server_config,			/* per-server init */
  NULL,				/* per-server merge */
  cas_directives,			/* command handler */
  cas_register_hooks			/* hook registration */
};
#else
module MODULE_VAR_EXPORT cas_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* init */
    NULL,			/* create per-dir config */
    NULL,			/* merge per-dir config */
    init_server_config,		/* create server config */
    NULL,			/* merge server config */
    cas_directives,		/* command table */
    NULL,			/* handler table */
    NULL,			/* translation */
    do_cas,			/* check/set authenticated user */
    authorize,			/* decide whether user is authorized */
    NULL,			/* access control by arbitrary criteria */
    NULL,			/* type checker */
    NULL,			/* "fixup" */
    NULL,			/* log handler */
#if MODULE_MAGIC_NUMBER >= 19970103
    NULL,			/* header parser */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
    init_child,			/* per-process initializer */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
    cleanup_child,		/* per-process exit/cleanup */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
    NULL,			/* post read_request handler */
#endif
};
#endif
